;;; ///////////////////////////////////////////////////////////////////
;;;
;;; CAPITOL 5 & 8
;;; 
;;; Reprezentare arbori
;;; fisier arb.lsp
;;;
;;; //////////////////////////////////////////////////////////////////

;;   Un arbore este reprezentat ca o colectie de noduri, fiecare nod fiind
;; legat la fiii sai. Descrierile nodurilor sunt tinute intr-o tabela hash
;; si sunt identificate de un identificator unic. Radacina este identificata
;; printr-un identificator special 'radacina'


;; ///// Un nod in arbore

(defstruct NOD
  id         ; identificator NOD 
  info       ; (regula-concluzie sau fapt)
  regula-id  ; id regula a carei concluzie este
  fii        ; descendenti
  )


(defun make-NOD-nou (info regula-id fii)
   ;; creeaza un NOD cu info si fii 
   ;; id este generat random  
   (make-NOD :id (gensym)
              :info info
              :regula-id regula-id
              :fii fii))

(defun constr-NOD (id info regula-id fii)
   ;; modifica un nod specificat prin id
   (make-NOD :id id
              :info info
              :regula-id regula-id
              :fii fii))

(defun obtine-NOD (NOD-id arbore)
   ;; Intoarce descrierea unui nod specificat prin id
   (let ((NOD (cdr (assoc NOD-id arbore :test #'equal))))
      (if (null NOD)
         (format t "~&Nod nul : ~a ~%" NOD-id)
         NOD)))

(defun mem-NOD (NOD arbore)
   ;; Memoreaza NOd in arbore
   (acons (NOD-id NOD) NOD arbore))

(defun sterge-NOD (NOD-id arbore)
   ;; Sterge NOD din arbore
   (remove (assoc NOD-id arbore) arbore))


;; /////// Functii auxiliare

(defun frunza-p (NOD-id arbore)
   (let ((NOD (obtine-NOD NOD-id arbore)))
      (if (null NOD)
         (format t "~&frunza-p : Null NOD - ~a ~%" NOD NOD-id)
         (null (NOD-fii NOD)))))

(defun radacina-p (NOD-id)
   (equal NOD-id 'radacina))

(defun adauga-fiu (id-tata nod arbore)
   ;; adauga un fiu la nodul id-tata
   ;(format t "~&adauga fiu ~a la ~a in ~a~%" (NOD-id NOD) id-tata arbore)
   (let ((nod-tata (obtine-NOD id-tata arbore)))
      (mem-NOD NOD                        
        (mem-NOD (constr-NOD id-tata 
                    (NOD-info nod-tata) 
                    (NOD-regula-id nod-tata)
                    (append (list (NOD-id NOD)) (NOD-fii nod-tata)))
          (sterge-NOD id-tata arbore))))) 


(defun sterge-fiu (id-tata fiu-id arbore)
   ;; Sterge un fiu din lista de fii a tatalui si din arbore. 
   (let ((nod-tata (obtine-NOD id-tata arbore)))
      (sterge-NOD fiu-id
        (mem-NOD (constr-NOD id-tata 
                    (NOD-info nod-tata)
                    (NOD-regula-id nod-tata) 
                    (remove fiu-id (NOD-fii nod-tata)))
          (sterge-NOD id-tata arbore)))))


(defun act-info (NOD-id arbore info)
   ;; actualizeaza info din NOd-id 
   (let ((NOD (obtine-NOD NOD-id arbore)))
      (mem-NOD (constr-NOD NOD-id
                               info
                               (NOD-regula-id NOD)
                               (NOD-fii NOD)) 
        (sterge-NOD NOD-id arbore))))
 

(defun act-regula-id (NOD-id arbore regula-id)
   ;; actualizeaza id regula al regulii cu concluzia in nodul curent
   (let ((NOD (obtine-NOD NOD-id arbore)))
      (mem-NOD (constr-NOD NOD-id
                               (NOD-info NOD)
                               regula-id
                               (NOD-fii NOD)) 
        (sterge-NOD NOD-id arbore))))
 
(defun tip-NOD (NOD)
   ;; tipareste NOD
   (if (not (NOD-p NOD))
      (format t "~&Nu este un NOD!~%")
      (format t "~&NOD : ~a ~% ~a~% ~a~% fii ~a~%" (NOD-id NOD)
         (NOD-info NOD) (NOD-regula-id NOD) (NOD-fii NOD))))




;; ////////////////////  Functii pe arbore 
  

(defun frunze (arbore)
   "Intoarce lista frunzelor din arbore"
   (frunze-aux arbore 'radacina ()))

(defun frunze-aux (arbore NOD-id lista-frunze)
   "Functie auxiliara pentru frunze"
   (if (frunza-p NOD-id arbore)
      (cons NOD-id lista-frunze)
      (dolist (fiu-id (NOD-fii (obtine-NOD NOD-id arbore)) lista-frunze)
         (setq lista-frunze (frunze-aux arbore fiu-id lista-frunze)))))


(defun make-arbore (radacina-info radacina-fii)
   ;; Creeaza un nou arbore continind radacina-info si radacina-fii
   (let ((arbore nil))    
      (mem-NOD (make-NOD :id 'radacina         
                           :info radacina-info  
                           :regula-id nil
                           :fii radacina-fii)  
        arbore)))                        


(defun dim (arbore)
   "Intoarce dimensiunea unui arbore (numarul de implicatii)"
   (dim-aux 'radacina 0 arbore))

(defun dim-aux (NOD-id rezultat arbore)
   "Functie auxiliara pentru dim"
   ;; Traverseaza depth first arbore cu radacina NOd-id
   ;; si calculeaza dimensiunea
   (if (frunza-p NOD-id arbore)
      rezultat
      (let ((rezultat 0)
            (NOD (obtine-NOD NOD-id arbore)))
         (dolist (fiu-id (NOD-fii NOD) (+ rezultat rezultat))
            (setq rezultat (+ rezultat (dim-aux fiu-id (+ 1 rezultat) arbore)))))))


                   
(defun dup-arbore (arbore)
   "Duplica arbore"
   (copy-alist arbore))    
     

(defun tip-arbore (arbore NOD-id)
   ;; Tipareste depth first arbore cu radacina NOD-id
   ;(format t "~& arbore : ~a~%" arbore)
   (let* ((NOD (obtine-NOD NOD-id arbore))
          (fii (NOD-fii NOD)))
      (if (null fii)
         (tip-NOD NOD)
         (progn 
           (tip-NOD NOD)
           (dolist (fiu fii nil) 
              (tip-arbore arbore fiu))))))
