
;;//////////////////////////////////////////////////////////////////////
;;
;; CAPITOLUL 8
;;
;;
;; Algoritmul ID3
;; Fisier id3.lsp
;;/////////////////////////////////////////////////////////////////////



;///////////////  Structuri de date   ////////////////////////

(defstruct arb-decizie 
  "Un arbore de decizie"
  nume-test
  test
  ramuri)

(defstruct frunza
  "O frunza a unui astfel de arbore"
  valoare)

(defstruct proprietate 
  nume   ;doar pentru user
  test   ;functie care intoarce o valoare a unei proprietati
  valori ;lista tuturor valorilor ce pot fi intoarse de test
)

(defstruct set-exemple
  "Set de exemple de antrenare"
  instante   ;lista clasificarilor de instante cunoscute
  proprietati  ;lista de structuri de tip proprietate
  clasificare   ;clasificarea pe care trebuie sa o invete
  nr        ;nr de exemple
  informatie ;continutul de informatie al setului de exemple
 )


(defstruct partitie
  "O partitie a exemplelor in subprobleme, folosind o anumita proprietate"
  nume-test   ;numele testului
  test        ;proprietatea folosita pentru crearea partitiei
  componente  ;subproblemele partitiei (a-lista: chei-valorile testului selectat, data-instanta
              ; a lui set-exemple
  castig-info   ;castigul in informatie ce rezulta in urma folosirii test-ului ca nod
  )


;//////////////  Functii accesor  /////////////////////////

(defun istorie-credit (obiect)
   "Intoarce proprietatea istorie-credit asociata exemplului obiect"
   (cdr (assoc 'istorie-credit obiect :test #'equal)))

(defun datorie (obiect)
   "Intoarce proprietatea datorie asociata exemplului obiect"
   (cdr (assoc 'datorie obiect :test #'equal)))

(defun garantie (obiect)
   "Intoarce proprietatea garantie asociata exemplului obiect"
   (cdr (assoc 'garantie obiect :test #'equal)))

(defun venit (obiect)
   "Intoarce proprietatea venit asociata exemplului obiect"
   (cdr (assoc 'venit obiect :test #'equal)))

(defun risc (obiect)
   "Intoarce proprietatea risc asociata exemplului obiect"
   ;; OBS. Aceasta proprietate este chiar clasificarea exemplului respectiv
   (cdr (assoc 'risc obiect :test #'equal)))



;///////////////// Constructia inductiva a arborelui ////////////////

(defun ind-arbore (ex-invatare)
   "Construieste inductiv arborele ce clasifica ex-invatare"
   (cond
         ;cazul 1 - set exemple vid
         ((null (set-exemple-instante ex-invatare))
          (make-frunza :valoare "imposibil de clasificat: nici un exemplu"))
         
         ;cazul 2 - toate testele au fost folosite 
         ((null (set-exemple-proprietati ex-invatare))
          (make-frunza :valoare (list-clase ex-invatare)))
         
         ;cazul 3 - toate exemplele in aceeasi clasa 
         ((zerop (set-exemple-informatie ex-invatare))
          (make-frunza :valoare (funcall
                              (proprietate-test (set-exemple-clasificare ex-invatare))
                              (car (set-exemple-instante ex-invatare)))))
         
         ;cazul 4 - selectioneaza testul si apel recursiv
         (t (let ((part (alege-partitie (gen-partitii ex-invatare))))
               (make-arb-decizie
                :nume-test (partitie-nume-test part)
                :test (partitie-test part)
                :ramuri (mapcar #'(lambda (x) (cons (car x) (ind-arbore (cdr x))))
                            (partitie-componente part)))))))


(defun gen-partitii (ex-invatare)
   "Genereaza toate partitiile instantelor lui ex-invatare"
   ;;Fiecare partitite e creata folosind o proprietate diferita.
   (mapcar #'(lambda (x) (partitie ex-invatare x))
     (set-exemple-proprietati ex-invatare)))

        
;alege-partitie cauta intr-o lista de candidati si o alege pe cel cu cel mai mare castig in 
; informatie

(defun alege-partitie (candidati)
   "Cauta in lista candidati si il alege pe cel 
cu cel mai mare castig in informatie "
   (cond ((null candidati) nil)
         ((= (list-length candidati) 1) (car candidati))
         (t (let ((maxim (alege-partitie (cdr candidati))))
               (if (> (partitie-castig-info (car candidati) (partitie-castig-info-maxim)))
                  (car candidati)
                  maxim)))))


(defun partitie (radacina proprietate)
   "Partitioneaza arborele radacina folosind proprietate"
   ;;Intoarce o instanta a structurii partitie.
   (let ((parti (mapcar #'(lambda (x) (cons x (make-set-exemple)))
                  (proprietate-valori proprietate))))
      (dolist (instanta (set-exemple-instante radacina))
         (push instanta (set-exemple-instante
                         (cdr (assoc (funcall (proprietate-test proprietate) instanta) parti)))))
      (mapcar #'(lambda (x)
                  (let ((ex (cdr x)))
                     (setf (set-exemple-proprietati ex)
                           (remove proprietate (set-exemple-proprietati radacina)))
                     (setf (set-exemple-clasificare ex)
                           (set-exemple-clasificare radacina))
                     (setf (set-exemple-nr ex)
                           (list-length (set-exemple-instante ex)))
                     (setf (set-exemple-informatie ex)
                           (calc-informatie
                            (set-exemple-instante ex)
                            (set-exemple-clasificare radacina)))))
        parti)
      (make-partitie 
       :nume-test (proprietate-nume proprietate)
       :test (proprietate-test proprietate)
       :componente parti
       :castig-info (calc-castig-info radacina parti))))
 
 

;////////////////  Functii axuiliare  /////////////////////////////


(defun list-clase (ex-invatare)
   "Creeaza un nod pentru o clasificare ambigua"
   ;;folosita in cazul 2 al lui ind-arbore
   (do
       ((clase (proprietate-valori (set-exemple-clasificare ex-invatare)) (cdr clase))
        (clasificare (proprietate-test (set-exemple-clasificare ex-invatare))) clase-prezente)
       ((null clase) clase-prezente)
      (if (member (car clase) (set-exemple-instante ex-invatare)
            :test #'(lambda (x y) (equal x (funcall clasificare y))))
         (push (car clase) clase-prezente))))



;////////////////  Functii pentru calculul informatiei ////////////////

(defun calc-informatie (exemple clasificare)
   "Determina continutul de informatie al unei liste de exemple"
   ;;Se calculeaza numarul instantelor din fiecare clasa si proportia din setul 
   ;;de antrenare care apartine fiecarei clase. Presupunand ca aceasta este 
   ;;probabilitatea ca un obiect sa apartina unei clase , se calculeaza continutul
   ;; de informatie din exemple, folosind def. lui Shannon
   (let ((index-clase 
          (mapcar #'(lambda (x) (cons x 0))
            (proprietate-valori clasificare)))
         (nr 0))
      ;numara instantele din fiecare clasa
      (dolist (instanta exemple)
         (incf nr)
         (incf (cdr (assoc (funcall (proprietate-test clasificare) instanta)
                      index-clase))))
      
      ;calc continutul de info al exemple
      (suma #'(lambda (x) (if (= (cdr x) 0)
                            0
                            (* -1
                               (/ (cdr x) nr)
                               (log (/ (cdr x) nr) 2))))
               index-clase)))


;calc-castig-info calc cresterea in info a unei partitii prin scaderea sumei ponderate a info in 
; componentele sale  din informatia din setul de exemple parinte

(defun calc-castig-info (radacina parti)
   "Calculeaza cresterea in info a unei partitii"
   ;;se scade suma ponderata a informatiei din componentele sale
   ;;din informatia din setul de exemple parinte
   (- (set-exemple-informatie radacina)
      (suma #'(lambda (x) (* (set-exemple-informatie (cdr x))
                            (/ (set-exemple-nr (cdr x))
                               (set-exemple-nr radacina))))
        parti)))

(defun suma (f lista-nr)
   "Calc suma aplicarii lui f pe fiecare din numerele din lista-nr"
   (apply '+ (mapcar f lista-nr)))



;///////////////   Functia de clasificare   //////////////////////

(defun clasifica (instanta arbore)
   "Determina clsificarea instantei pe baza arborelui"
   (if (frunza-p arbore)
      (frunza-valoare arbore)
      (clasifica instanta 
       (cdr (assoc (funcall (arb-decizie-test arbore) instanta)
              (arb-decizie-ramuri arbore))))))
               


;;////////////   Exemple  /////////////////////////////////

;Set exemple de invatare

(setq exemple
  '(((risc . mare) (istorie-credit . proasta) (datorie . mare) (garantie . nimic) (venit . 0-15k))
    ((risc . mare) (istorie-credit . necunoscut) (datorie . mare) (garantie . nimic) (venit . 15-35k))
    ((risc . mediu) (istorie-credit . necunoscut) (datorie . mica) (garantie . nimic) (venit . 15-35k))
    ((risc . mare) (istorie-credit . necunoscut) (datorie . mica) (garantie . nimic) (venit . 0-15k))
    ((risc . mic) (istorie-credit . necunoscut) (datorie . mica) (garantie . nimic) (venit . peste-35k))
    ((risc . mic) (istorie-credit . necunoscut) (datorie . mica) (garantie . adecvat) (venit . peste-35k))
    ((risc . mare) (istorie-credit . proasta) (datorie . mica) (garantie . nimic) (venit . 0-15k))
    ((risc . mediu) (istorie-credit . proasta) (datorie . mica) (garantie . adecvat) (venit . peste-35k))
    ((risc . mic) (istorie-credit . buna) (datorie . mica) (garantie . nimic) (venit . peste-35k))
    ((risc . mic) (istorie-credit . buna) (datorie . mare) (garantie . adecvat) (venit . peste-35k))
    ((risc . mare) (istorie-credit . buna) (datorie . mare) (garantie . nimic) (venit . 0-15k))
    ((risc . mediu) (istorie-credit . buna) (datorie . mare) (garantie . nimic) (venit . 15-35k))
    ((risc . mic) (istorie-credit . buna) (datorie . mare) (garantie . nimic) (venit . peste-35k))))
 

    
;;Instanta de test

(setq test-instanta
  '((istorie-credit . proasta) (datorie . mare) (garantie . nimic) (venit . 15-35k)))


;Lista testelor

(setq teste
  (list
    (make-proprietate 
     :nume 'istorie-credit
     :test #'istorie-credit
     :valori '(buna proasta necunoscut))
    (make-proprietate
     :nume 'datorie
     :test #'datorie
     :valori '(mare mica))
    (make-proprietate
     :nume 'garantie
     :test #'garantie
     :valori '(nimic adecvat))
    (make-proprietate
     :nume 'venit
     :test #'venit
     :valori '(0-15k 15-35k peste-35k))))


;Proprietatea ce trebuie invatata

(setq clasificare
  (make-proprietate
   :nume 'risc
   :test #'risc
   :valori '(mare mediu mic)))


;Setul de exemple de invatare

(setq credit-exemple
  (make-set-exemple
   :instante exemple
   :proprietati teste
   :clasificare clasificare
   :nr (list-length exemple)
   :informatie (calc-informatie exemple clasificare)))
 



;;//////////////////// Exemplu apel ////////////////////////


;Constructia arborelui de decizie pe baze exemplelor

(setq credit-arbore (ind-arbore credit-exemple))


;Clasificarea exemplului de test

(clasifica test-instanta credit-arbore)