;; ////////////////////////////////////////////////////////////////
;;
;; CAPITOL 8
;;     
;; Gasirea celei mai simple explicatii pentru o concluzie
;; fisier expl.lsp
;;
;; /////////////////////////////////////////////////////////////////


(defun gaseste-expl (concl)
   "Gaseste cea mai simpla explicatie pentru concl"
   ;; O explicatie e reprezentata printr-un arbore de justificare
   ;; si un set de substitutii. Functia filtreaza rezultatul lui 
   ;; rezolva-init pentru a gasi cea mai simpla explicatie,
   ;; presupusa a fi si cea mai plauzibila (Occam)
   (gaseste-cea-mai-simpla-expl () (rezolva-init concl)))


(defun gaseste-cea-mai-simpla-expl (sol-ant stream-solutie)
   "Filtreaza stream solutie pentru a gasi cea mai simpla explicatie"
   ;; Relatia de ordine pe arborii de justificare e data de dimensiunea lor
   ;; dimensiune (arbore) = numar de nivele (implicatii)
   (if (stream-vid-p stream-solutie)
      sol-ant
      (let* ((sol-curenta (head-stream stream-solutie))
             (arb-curent (solutie-arb sol-curenta))
             (subst-curenta (solutie-subst sol-curenta)))
         (if (null sol-ant)
            (gaseste-cea-mai-simpla-expl (list subst-curenta arb-curent)
              (tail-stream stream-solutie))
            (let ((arb-ant (solutie-arb sol-ant))
                  (subst-ant (solutie-subst sol-ant))) 
               (if (< (dim arb-curent) (dim arb-ant))
                  (gaseste-cea-mai-simpla-expl (list subst-curenta arb-curent)
                    (tail-stream stream-solutie))
                  (gaseste-cea-mai-simpla-expl sol-ant (tail-stream stream-solutie)))))))) 


;;/////////////////////////////////////////////////////////////////////////
;;
;; Exemplu apel
;;
;; Functia care gaseste cea mai simpla explicatie a unei concluzii este gaseste-expl,
;; ce primeste ca argument concluzia de demonstrat. De exemplu, pentru baza de date
;; data (BC.txt), apelul
;;
;; (gaseste-expl '(cana obj1))
;;
;; va gasi cea mai simpla explicatie a faptului ca obj1 este o cana si va intoarce 
;; aceasta explicatie sub forma de arbore de explicatie impreuna cu sirul substitutiilor
;; sub care a fost demonstrata concluzia. Structura de date intoarsa de gaseste-expl este
;; de tipul solutie (un element al streamului de solutii intors de rezolva-init)