;;; /////////////////////////////////////////////////////////////
;;;
;;; CAPITOL 5 & 8
;;;
;;; Motor de inferenta backward chaining
;;; fisier engine.lsp
;;;
;;; ////////////////////////////////////////////////////////////


(defvar *ASERTIUNE-CURENTA* ())

(defvar *lista-nume* ())

          
         
;; ////// Functii auxiliare

(defun concluzie-conj-p (concl)
   ;; Testeaza daca o concluzie este conjunctiva
   (equal (nth 0 concl) 'and))

(defun corp (concl)
   ;; Intoarce corpul unei concluzii
   (cdr concl))

(defun filtrare-prin-concl-conj (concl stream-solutie nod-id)
   ;; Trece streamul solutie printr-o secventa de concluzii,
   ;; eliminand acele substitutii care nu sunt consistente cu
   ;; o anumita concluzie.
   ;; Construieste in paralel si structura de explicatie ce 
   ;; justifica conaluzia identificata de nod-id
   (if (null concl)
      stream-solutie
      (filtrare-prin-concl-conj (cdr concl) 
       (filtrare-prin-concluzie (car concl) stream-solutie nod-id)
       nod-id)))

   
(defun filtrare-prin-concluzie (concl stream-solutie nod-id) 
   ;; Filtreaza stream de substitutii printr-o concluzie
   (if (stream-vid-p stream-solutie)
      (make-stream-vid)
      (let* ((solutie (head-stream stream-solutie))
             (subst (solutie-subst solutie))
             (arbore (solutie-arbore solutie)))
         (if (not (equal (NOD-info (obtine-nod nod-id arbore)) concl))
            (let* ((nod-nou (make-NOD-nou concl nil ()))
                   (arb-nou (adauga-fiu nod-id nod-nou arbore)))
               (combina-streamuri 
                 (rezolva concl subst arb-nou (nod-id nod-nou))
                 (filtrare-prin-concluzie concl (tail-stream stream-solutie) nod-id)))
            (combina-streamuri 
              (rezolva concl subst arbore nod-id)
              (filtrare-prin-concluzie concl (tail-stream stream-solutie) nod-id))))))

  


;; /////////////////// Functii de rezolvare /////////////////////////////////

(defun rezolva (concl subst arbore nod-id)
   ;; Primeste o concluzie si un set de substitutii si
   ;; gaseste toate solutiile consistente cu BC
   ;; Construieste arborele de explicatie ce justifica 
   ;; concluzia, si il plaseaza in arbore, in nodul 
   ;; identificat cu nod-id
   (if (concluzie-conj-p concl)
      (let ((nod-nou (make-nod-nou (car (corp concl)) () ()))) 
         (filtrare-prin-concl-conj (cdr (corp concl)) 
          (rezolva (car (corp concl)) subst 
            (adauga-fiu nod-id nod-nou arbore) (nod-id nod-nou)) 
          nod-id))
      (rezolva-concl-simpla concl subst arbore nod-id)))


(defun rezolva-concl-simpla (concl subst arbore nod-id)
   ;; rezolva o concl simpla 
   (declare (special *ASERTIUNE-CURENTA*))
   (if (not (equal (NOD-info (obtine-nod nod-id arbore)) concl))
      (let ((nod-nou (make-nod-nou concl () ())))
         (infer concl subst *ASERTIUNE-CURENTA*
           (adauga-fiu nod-id nod-nou arbore) (nod-id nod-nou)))
      (infer concl subst *ASERTIUNE-CURENTA* arbore nod-id)))    


(defun rezolva-regula (regula subst arbore nod-id)
   ;; gaseste toate solutiile pentru premizele regulii
   (setq arbore (act-regula-id nod-id arbore (regula-id regula)))
   (rezolva (premize regula) subst arbore nod-id))


(defun infer-din-fapte (concl stream-fapte arbore)
   (if (stream-vid-p stream-fapte)
      (make-stream-vid)
      (cons-stream
        (list (head-stream stream-fapte) arbore)
        (infer-din-fapte concl (tail-stream stream-fapte) arbore))))


(defun infer (concl subst asert-curenta arbore nod-id)
   ;; Infereaza concl din BC, sub un set de substitutii
   ;; intoarce un stream de substitutii 
   ;; Construieste structura de explcatie care justifica
   ;; concluzia, rezultatul fiind un arbore cu radacina nod-id
   (if (equal asert-curenta 'sfarsit-BC)
      (infer-din-fapte concl (gen-fapte concl subst) arbore)                       
      (let* ((asert (obtine-asert asert-curenta))    
             (res (unifica concl (concluzie asert) subst))
             (asert-urm (obtine-asert-urm asert-curenta)))
         (if (equal res 'insucces)
            (infer concl subst asert-urm arbore nod-id)                              
            (combina-streamuri                    
              (rezolva-regula asert res arbore nod-id)
              (infer concl subst asert-urm arbore nod-id)))))) 
  
    
(defun rezolva-init (concl)
   "rezolva concl. intorcand un stream solutie"
   (setq *ASERTIUNE-CURENTA* (first *LISTA-ASERTIUNI*))
   (rezolva concl nil (make-arbore concl ()) 'radacina))



;; ///////////////////////////////////////////////////////////////////////
;;   Exemplu de apel
;;
;; Motorul de inferenta se va apela cu functia rezolva-init cu argumentul
;; concluzia ce trebuie demonstrata. De exemplu, pentru baza de cunostinte
;; data (bc.txt), un apel ar putea fi:
;;
;; (rezolva-init '(cana obj1))
;;
;; Aceasta functie va intoarce un rezultat
;; mai complex (asa cum e descris in carte), de fapt un stream de substitutii si
;; arbore de explicatie. Pentru selectarea numai a substitutiilor din acest stream
;; se foloseste functia solutie-subst ce primeste ca argument streamul solutie
;; si intoarce substitutiile din el. De exemplu, apelul
;;
;; (stream-solutie-subst (rezolva-init '(cana obj1))
;;
;; va intoarce substitutiile sub care se poate demonstra din baza de cunostinte
;; concluzia (cana obj1). 
;; OBS. Aceasta lista de substitutii este una din solutiile gasite de sistem. Pentru
;; gasirea restului solutiilor se va "forta" streamul solutiile a carui evaluare a fost
;; intarziata (stream intors de rezolva-init). Astfel, apelul
;;
;; (stream-solutie-subst (tail-stream (rezolva-init '(cana obj1))))
;;
;; va intoarce alte posibile substitutii sub care se poate demonstra concluzia
;; In cazul prezentat prima solutie este si singura, deci apelul de mai sus va intoarce nil