;;/////////////////////////////////////////////////////////////////
;;
;; 6 Cunostinte structurate
;; 6.3 Distanta inferentiala
;; 6.3.2 Mostenirea bazata pe distanta inferentiala
;;
;;///////////////////////////////////////////////////////////////

;; "distinfer.lsp" - gaseste valorile sloturilor utilizand distanta 
;; inferentiala

(defvar *UNITATE*)

(defvar *SLOT*)

(defun uassoc (cheie lista-asociatii)
  (cond ((assoc cheie (rest lista-asociatii)))
        (t (cadr (rplacd (last lista-asociatii) 
(list (list cheie)))))))

(defun obtine-unitate (unitate)
  (when unitate
    (cond ((get unitate 'UNITATE))
          (t (setf (get unitate 'UNITATE) (list unitate))))))

(defun u-adauga (unitate slot fateta valoare)
  (cond 
    ((member valoare (u-obtine unitate slot fateta))
      nil)
    (t (push valoare
         (rest (uassoc fateta
                 (uassoc slot (obtine-unitate unitate)))))
       valoare)))

(defun u-obtine (unitate slot fateta)
  (rest (assoc fateta
           (rest (assoc slot
                    (rest (get unitate 'UNITATE)))))))

(defun u-sterge (unitate slot fateta valoare)
  (let* ((sloturi (obtine-unitate unitate))
         (fatete (assoc slot (rest sloturi)))
         (valori (assoc fateta (rest fatete))))
    (when (member valoare (rest valori))
      (setf (rest valori) (delete valoare (rest valori)))
      (unless (rest valori)
          (delete valori fatete))
      (unless (rest fatete)
          (delete fatete sloturi))
      t)))

(defun u-verifica (unitate slot fateta valoare)
  (when (member valoare (u-obtine unitate slot fateta))
    t))

(defun u-leaga (unitate1 unitate2 slot)
  (rplacd (uassoc slot (obtine-unitate unitate1))
          (rest (uassoc slot (obtine-unitate unitate2))))
  slot)

(defun u-obtine-vi (unitate slot)
  (cond ((u-obtine unitate slot 'VALOARE))
        ((u-obtine unitate slot 'VALOARE-IMPLICITA))))

(defun u-obtine-vmi (unitate slot)
  (let ((*UNITATE* unitate)
        (*SLOT* slot))
    (cond 
       ((u-obtine unitate slot 'VALOARE))
       ((delete nil (mapcar #'(lambda (metoda)
                                   (funcall metoda))
                       (u-obtine unitate slot 'METODA))))
       (t (u-obtine unitate slot 'VALOARE-IMPLICITA)))))

;; exemple de fatete METODA

(defun intreaba nil
(format t "~&Furnizati o valoare pentru slotul ~A.~A " *SLOT* *UNITATE*)
  (read))

(defun greutate-barbat nil
  (let ((rezultat (u-obtine-vi *UNITATE* 'Inaltime)))
    (when rezultat
      (* 30 (first rezultat)))))

(defun greutate-femeie nil
  (let ((rezultat (u-obtine-vi *UNITATE* 'Inaltime)))
    (when rezultat
      (* 20 (first rezultat)))))

(defun ierarhie (unitate)
  (do ((coada (list unitate) (rest coada))
        stramos unitati)
      ((null coada) (reverse unitati))
    (unless (member (first coada) unitati)
      (push (first coada) unitati))
    (setf stramos (u-obtine (first coada) 'AKO 'VALOARE))
    (setf coada (append coada stramos))))

(defun obtine-valoare-I (unitate slot)
  (do ((unitati (ierarhie unitate) (rest unitati))
       rezultat)
      ((cond ((null unitati))
             ((setf rezultat (u-obtine (first unitati) 
  			slot 'VALOARE))))
       rezultat)))

(defun  obtine-valoare-Z (unitate slot)
  (do* ((unitati (ierarhie unitate) (rest unitati))
        (unitate1 (first unitati) (first unitati))
        rezultat)
       ((cond ((null unitati))
              ((setf rezultat 
  (u-obtine-vmi unitate1 slot))))
        rezultat)))

(defun obtine-valoare-N (unitate slot)
  (let ((unitati (ierarhie unitate))
        rezultat)
    (do ((unitati1 unitati (rest unitati1)))
        ((cond ((null unitati1))
               ((setf rezultat (u-obtine (first unitati1) slot 'VALOARE))))))
    (cond (rezultat rezultat)
          ((do* ((unitati1 unitati (rest unitati1))
                 (*UNITATE* (first unitati1) (first unitati1))
                 (*SLOT* slot))
                ((cond ((null unitati1))
                       ((setf rezultat
                          (delete nil
                            (mapcar #'(lambda (metoda)
                                         (funcall metoda))
                              (u-obtine *UNITATE* slot
                                  'METODA))))))
                 t))
           (cond (rezultat rezultat)
                 ((do ((unitati1 unitati (rest unitati1)))
                      ((cond 
                         ((null unitati1))
                         ((setf rezultat (u-obtine (first unitati1)    
                                            slot 'VALOARE-IMPLICITA))))
                       t))
                  rezultat))))))

(defun u-adauga+ (unitate slot fateta valoare)
  (when (u-adauga unitate slot fateta valoare)
    (let ((*SLOT* slot))
      (mapc #'(lambda (stramos)
                (let ((*UNITATE* stramos))
                  (mapc #'(lambda (demon) (funcall demon))
                     (u-obtine stramos slot 'ADAUGARE?))))
            (ierarhie unitate)))
    valoare))

(defun u-sterge+ (unitate slot fateta valoare)
  (when (u-sterge unitate slot fateta valoare)
    (let ((*SLOT* slot))
      (mapc #'(lambda (stramos)
                (let ((*UNITATE* stramos))
                  (mapc #'(lambda (demon) (funcall demon))
                     (u-obtine stramos slot 'STERGERE?))))
            (ierarhie unitate)))
    valoare))

(defmacro defunitate (unitate &rest argumente)
  `(defunitate1 ',unitate ',argumente))

(defun defunitate1 (unitate argumente)
  (let (sloturi marcaj)
    (when (symbolp unitate)
       (cond ((symbolp (first argumente))
              (u-adauga unitate 'AKO 'VALOARE (first argumente))
              (setf sloturi (rest argumente) marcaj t))
             ((setf sloturi argumente)))
       (do* ((lista-sloturi sloturi (rest lista-sloturi))
             (slot (caar lista-sloturi) (caar lista-sloturi))
             (fatete (cdar lista-sloturi) (cdar lista-sloturi)))
            ((null lista-sloturi) unitate)
         (uassoc slot (obtine-unitate unitate))
         (do* ((lista-fatete fatete (cddr lista-fatete))
               (fateta (first lista-fatete) (first lista-fatete))
               (valoare (cadr lista-fatete) (cadr lista-fatete)))
              ((or (null lista-fatete) (null (cadr lista-fatete))))
           (if (member fateta '(METODA ADAUGARE? STERGERE?))
             (u-adauga unitate slot fateta (symbol-function valoare))
             (u-adauga unitate slot fateta valoare)))
         (unless (u-obtine unitate slot 'STRATEGIE)
           (u-adauga unitate slot 'STRATEGIE 'obtine-valoare-N))
         (unless (or marcaj (u-obtine unitate slot 'METODA))
           (u-adauga unitate slot 'METODA #'intreaba))))))


(defmacro definstanta (instanta unitate &rest argumente)
  `(definstanta1 ',instanta ',unitate ',argumente))

;; la constructia instantei nu se mai cauta in ierarhie sloturi mostenite (
;; pentru functionarea corecta a testelor pentru distanta inferentiala
(defun definstanta1 (instanta unitate argumente)
  (let ((stramosi (ierarhie unitate)))
    (u-adauga instanta 'ISA 'VALOARE unitate)
    (do* ((lista-sloturi argumente (cddr lista-sloturi))
          (slot (first lista-sloturi) 
                (first lista-sloturi))
         (valoare (cadr lista-sloturi)
                  (cadr lista-sloturi)))
         ((or (null lista-sloturi) (null valoare)))
      (u-adauga instanta slot 'VALOARE valoare))))
   
;; macro-uri ce permit folosirea mai eleganta a 
;; unitatilor. 

(defmacro adds (obiect slot fateta valoare)
  `(u-adauga ',obiect ',slot ',fateta ',valoare))

(defmacro gets (obiect slot fateta)
  `(u-obtine ',obiect ',slot ',fateta))

(defmacro examineaza (obiect)
  `(get ',obiect 'UNITATE))

;;=======================================================================
;; Implementarea distantei inferentiale

(defun lista-ISA (unitate)
   (append (list unitate) (u-obtine unitate 'ISA 'VALOARE)))

;;(defun lista-ISA (unitate)
;;   (u-obtine unitate 'ISA 'VALOARE))

(defun lista-candidati (unitate slot)
   (do ((coada (lista-ISA unitate) (rest coada))
        candidati vizitate stramos)
       ((null coada) (reverse candidati))
      (unless (member (first coada) vizitate)
         (progn
           (push (first coada) vizitate)
           (if (u-obtine (first coada) slot 'VALOARE)
              (push (first coada) candidati)
              (progn 
                (setf stramos (u-obtine (first coada) 'AKO 'VALOARE))
                (setf coada (append coada stramos))))))))

(defun distanta-inferentiala (unitate slot)
   (let ((candidati (lista-candidati unitate slot)))
    (do* ((cand candidati (rest cand))
         (prim (first cand) (first cand)))
         ((null cand) candidati)
      (do* ((superclase (rest (ierarhie prim))
                        (rest superclase))
            (curent (first superclase)
                    (first superclase)))
           ((null superclase))
         (if (member curent candidati)
            (progn
              (setf candidati (remove curent candidati))
              (setf cand (remove curent cand))))))))

(defun u-obtine-inf (unitate slot)
   (let ((candidati (distanta-inferentiala unitate slot)))
     (print candidati)
     (cond 
       ((null candidati) nil )
       ((= 1 (length candidati)) 
        (u-obtine (first candidati) slot 'VALOARE))
       (t nil))))

;; Exemplu de utilizare

(defunitate Cactus (inteapa VALOARE 'da))
(defunitate CactusAltoit Cactus (inteapa VALOARE 'nu))
(defunitate CactusRosu CactusAltoit)
(defunitate Gymnocalicium CactusRosu)
(defunitate CactusDinSera Cactus)
(definstanta C10 CactusDinSera)
(obtine-unitate 'C10)
(u-adauga 'C10 'ISA 'VALOARE 'Gymnocalicium)
(obtine-unitate 'C10)
(u-obtine-inf 'C10 'inteapa)

;; Observatie. Implementarea ierarhiei trebuie modificata astfel incat sa 
;; foloseasca distanta inferentiala la deducerea valorii unui slot. De asemenea 
;; trebuie modificate definitiile claselor si instantelor astfel incat sa 
;; permita mostenirea multipla.


