;;////////////////////////////////////////////////////////////
;;
;;3 Strategii de cautare
;;3.3 Rezolvarea problemelor utilizand strategii de cautare
;;3.3.3 Problema mozaicului 
;;
;;/////////////////////////////////////////////////////////////

;; "mozaic.lsp" - rezolva problema mozaicului cu 8 cifre, utilizand strategiile 
;; de cautare in adancime, pe nivel, si A* cu doua functii euristice

;; Reprezentare a starii pentru problema mozicului cu 8 cifre.
;; (pozitie-blanc ((poz1.val1)(poz2.val2) ... (poz9.val9)))


(defparameter *FINAL* '(5 ((1 . 1) (2 . 2) (3 . 3) (4 . 8) (5 . blanc) (6 . 4)
                           (7 . 7) (8 . 6) (9 . 5)))) 

(defparameter *INITIAL* '(1 ((1 . blanc) (2 . 2) (3 . 3) (4 . 1) (5 . 8) (6 . 4) 
                             (7 . 7) (8 . 6) (9 . 5))))

(defparameter *STARI-INITIALE* 
  '((1 . (1 ((1 . blanc)(2 . 2)(3 . 3)(4 . 1)(5 . 8)(6 . 4)(7 . 7)
             (8 . 6)(9 . 5))))
    (2 . (3 ((1 . 2)(2 . 3)(3 . blanc)(4 . 1)(5 . 8)(6 . 4)(7 . 7)
             (8 . 6)(9 . 5))))
    (3 . (5 ((1 . 1)(2 . 3)(3 . 4)(4 . 8)(5 . blanc)(6 . 2)(7 . 7)
             (8 . 6)(9 . 5))))
    (4 . (8 ((1 . 2)(2 . 8)(3 . 3)(4 . 1)(5 . 6)(6 . 4)(7 . 7)
             (8 . blanc)(9 . 5))))
    (5 . (5 ((1 . 2)(2 . 1)(3 . 6)(4 . 4)(5 . blanc)(6 . 8)(7 . 7)
             (8 . 5)(9 . 3))))
    (6 . (9 ((1 . 6)(2 . 2)(3 . 7)(4 . 1)(5 . 5)(6 . 3)(7 . 8)
             (8 . 4)(9 . blanc))))
    (7 . (2 ((1 . 8)(2 . blanc)(3 . 4)(4 . 1)(5 . 2)(6 . 3)(7 . 7)
             (8 . 6)(9 . 5))))) )

(defparameter *SOLUTII* nil)

(defun init (numar-stare)
   (let ((pereche (assoc numar-stare *STARI-INITIALE*)))
      (if (equal pereche nil)
         (format t "~& Nu exista o stare initial cu numarul: ~a" numar-stare)
         (defparameter *INITIAL* (cdr pereche)))))

(init 1)

;;========================================================================
;;========================================================================

(defun estescop (x) (equal x *FINAL*))

(defun mutare-posibila (poz-blanc directie)
   (cond ((and (equal directie 'stanga) (member (mod poz-blanc 3) '(2 0))) t)
         ((and (equal directie 'dreapta) (> (mod poz-blanc 3) 0)) t)
         ((and (equal directie 'sus) (> poz-blanc 3)) t)
         ((and (equal directie 'jos) (< poz-blanc 7)) t)
         (t nil)))

(defun modifica-alista (alista poz1 poz2)
   (let ((val1 (cdr (assoc poz1 alista)))
         (val2 (cdr (assoc poz2 alista)))
         (rezultat (copy-alist alista)))
      (setf (cdr (assoc poz1 rezultat)) val2)
      (setf (cdr (assoc poz2 rezultat)) val1)
      rezultat))

(defun muta (poz-blanc a-lista directie)
   (cond ((equal directie 'stanga)
          (list (1- poz-blanc)
                (modifica-alista a-lista poz-blanc (1- poz-blanc))))
         ((equal directie 'dreapta)
          (list (1+ poz-blanc)
                (modifica-alista a-lista (1+ poz-blanc) poz-blanc)))
         ((equal directie 'sus)
          (list (- poz-blanc 3) (modifica-alista a-lista (- poz-blanc 3) 
                                  poz-blanc)))
         ((equal directie 'jos)
          (list (+ poz-blanc 3) (modifica-alista a-lista poz-blanc (+ poz-blanc 
                                                                 3))))
         (t nil)))

(defun succf (stare)
   (let ((poz-blanc (first stare)) (a-lista (cadr stare)) succ-list)
      (dolist (directie '(stanga dreapta sus jos) (reverse succ-list))
         (if (mutare-posibila poz-blanc directie)
            (progn
             (setf succ-list (cons (muta poz-blanc a-lista directie) 
                              succ-list)))))))

;;========================================================================
;;========================================================================

(defun cautgraf (FRONTIERA scop succesori strategie TERITORIU)
  (format t "~& CAUT: ~a" (first FRONTIERA))
  (print "=====================================")
  (cond ((null FRONTIERA) nil)
      ((funcall scop (first FRONTIERA)) (first FRONTIERA))
      (t (cautgraf
            (funcall strategie (expand FRONTIERA succesori TERITORIU)
                               (rest FRONTIERA))
            scop succesori strategie
            (adaug (first FRONTIERA) TERITORIU)))))

(defun expand (FRONTIERA succesori TERITORIU)
(remove-if
    #'(lambda (stare)
              (or (member stare FRONTIERA :test #'equal)
                  (member stare TERITORIU :test #'equal)))
    (funcall succesori (first FRONTIERA))))

(defun adaug (stare TERITORIU)
(if (member stare TERITORIU :test #'equal) TERITORIU (cons stare TERITORIU)))

(defun la_urma (x y) (append y x))

(defun cautgraf-lat (init scop succesori)
(cautgraf (list init) scop succesori #'la_urma nil))

(defun cautgraf-adinc (init scop succesori)
(cautgraf (list init) scop succesori #'append nil))

;;Exemple de apel
(cautgraf-adinc '(3 ((1 . 2)(2 . 3)(3 . blanc)(4 . 1)(5 . 8)(6 . 4)(7 . 7)
                     (8 . 6)(9 . 5))) #'estescop #'succf)
(cautgraf-lat *INITIAL* #'estescop #'succf)
(cautgraf-adinc *INITIAL* #'estescop #'succf)

;;=========================================================================
;;=========================================================================

(defstruct cale 
                stare
                (predecesor nil)) 

(defun expand-cale (FRONTIERA succesori TERITORIU test-st)
(let ((curenta (first FRONTIERA)))
 (mapcar #'(lambda (noua)
              (make-cale :stare noua  :predecesor  curenta))
         (remove-if #'(lambda (st)
                         (or (funcall test-st st FRONTIERA)
                             (funcall test-st st TERITORIU)))
                    (funcall succesori (cale-stare curenta))))))

(defun afis-cale (stare)
(do ((curenta stare (cale-predecesor curenta)))
    ((null (cale-predecesor curenta))
           (format t "~& STARE: ~a" (cale-stare curenta)))
    (format t "~& STARE: ~a" (cale-stare curenta))))

(defun scopcale (x)
(equal (cale-stare x) *FINAL*))

(defun estein (elem lista)
"Parametrul actual al lui test-st"
(some #'(lambda (x)
                (equal elem (cale-stare x)))
      lista))

(defun cautgraf (FRONTIERA scop succesori strategie TERITORIU test-st acumulator)
  ;(format t "~& CAUT: ~a" (first FRONTIERA))
  (cond ((null FRONTIERA) nil)
      ((funcall scop (first FRONTIERA)) (print (lungime-cale (first FRONTIERA) 0))
                                   (print acumulator))
      (t (cautgraf
            (funcall strategie (expand-cale FRONTIERA succesori TERITORIU test-st)
                               (rest FRONTIERA))
            scop succesori strategie
            (adaug (first FRONTIERA) TERITORIU test-st) test-st (1+ acumulator)))))

(defun lungime-cale (cale acumulator)
   (if (equal (cale-predecesor cale) nil)
      acumulator
      (lungime-cale (cale-predecesor cale) (1+ acumulator))))
   

(defun adaug (stare TERITORIU test-st)
(if (funcall test-st stare TERITORIU) TERITORIU (cons stare TERITORIU)))

(defun la_urma (x y) (append y x))

(defun cautgraf-lat (init scop succesori)
(cautgraf (list init) scop succesori #'la_urma nil #'estein 0))

(defun cautgraf-adinc (init scop succesori)
(cautgraf (list init) scop succesori #'append nil #'estein 0))

;;Exemplu de apel
(cautgraf-lat (make-cale :stare *INITIAL*) #'scopcale #'succf)

;;=========================================================================
;;=========================================================================

(defstruct cale stare g h f (predec nil))

(defun insert (x lista-ord)
   (setf (cale-f x) (+ (cale-g x) (cale-h x)))
   (cond ((null lista-ord) (list x))
          ((<= (cale-f x) (cale-f (first lista-ord)))
              (cons x lista-ord))
       (t (cons (first lista-ord) (insert x (rest lista-ord))))))

(defun succf* (x fct-h fct-g)
   (let ((xx (cale-stare x)))
        (mapcar #'(lambda (y) (make-cale :stare y :g (funcall fct-g y x) 
                                                  :h (funcall fct-h y) :predec x))
              (succf xx))))              

(defun estescop* (x) (equal (cale-stare x) *FINAL*))
   
(defun test-st* (stare FRONTIERA)
   (labels ((test (x lst1 lst2) 
               (cond ((null lst2) (values nil lst1 nil lst2))
                     ((equal x (cale-stare (first lst2))) 
                         (values t lst1 (first lst2) (rest lst2)))
                   (t (let (flag l1 e l2)
                         (multiple-value-setq (flag l1 e l2)
                                  (test x (append lst1 (list (first lst2)))
                                          (rest lst2)))
                         (values flag l1 e l2)))))) 
     (test (cale-stare stare) nil FRONTIERA)))

(defun lungime-cale (cale acumulator)
   (if (equal (cale-predec cale) nil)
      acumulator
      (lungime-cale (cale-predec cale) (1+ acumulator))))
   

(defun cautcale (FRONTIERA scop succesori TERITORIU test-st fct-h fct-g acumulator)
;;(format t"~& CAUT: ~a" (first FRONTIERA))
(cond 
 ((null FRONTIERA) nil)
 ((funcall scop (first FRONTIERA)) (print (first FRONTIERA))
                              (print acumulator)
                              (print (lungime-cale (first FRONTIERA) 0)))
 (t 
  (let ((succlist (funcall succesori (first FRONTIERA) fct-h fct-g)))
   ;;(print (cale-g (first FRONTIERA)))
   (setf TERITORIU (insert (first FRONTIERA) TERITORIU))      
   (setf FRONTIERA (rest FRONTIERA))
   (do ((crt succlist (rest crt)) )
         ((null crt) nil)
      (let (flag l1 elem l2)
         (multiple-value-setq (flag l1 elem l2) 
                              (funcall test-st (first crt) FRONTIERA))
         (if flag
            (if (< (cale-g (first crt)) (cale-g elem))
               (setf FRONTIERA (insert (first crt)
                                  (append l1 l2))))
            (progn
             (multiple-value-setq (flag l1 elem l2) 
                                  (funcall test-st (first crt) TERITORIU))
             (if flag
                 (if (< (cale-g (first crt)) (cale-g elem))
                    (progn 
                      (setf FRONTIERA (insert (first crt) FRONTIERA))
                      (setf TERITORIU (append l1 l2))))
                 (setf FRONTIERA (insert (first crt) FRONTIERA)))))))
    (cautcale FRONTIERA scop succesori TERITORIU test-st fct-h fct-g (1+ acumulator))))))



(defun g (stare-noua cale-veche)
   (+ 1 (cale-g cale-veche)))

(defun h (stare)
   (let ((poz-blanc1 (car stare)) (a-lista1 (cadr stare))
         (a-lista2 (cadr *FINAL*)) (rezultat 0))
      (do* ((al1 a-lista1 (rest al1))
            (al2 a-lista2 (rest al2))
            (pereche1 (first al1) (first al1))
            (pereche2 (first al2) (first al2))
            (key1 (car pereche1) (car pereche1))
            (key2 (car pereche2) (car pereche2))
            (val1 (cdr pereche1) (cdr pereche1))
            (val2 (cdr pereche2) (cdr pereche2)))
           ((null al1) rezultat)
         (if (not (equal key1 poz-blanc1))
            (if (not (equal val1 val2))
               (setf rezultat (1+ rezultat)))))))

(defun h1 (stare)
   (let ((poz-blanc1 (car stare)) (a-lista1 (cadr stare))
         (a-lista2 (cadr *FINAL*)) (rezultat 0))
      (mapcar 
        #'(lambda (x y) (if (not (equal (car x) poz-blanc1))
                           (if (not (equal (cdr x) (cdr y)))
                              (setf rezultat (1+ rezultat)))))
        a-lista1 a-lista2)
      rezultat))



(defun h2 (stare)
   (let ((poz-blanc1 (car stare)) (a-lista1 (cadr stare))
         (a-lista2 (cadr *FINAL*)) (rezultat 0))
      (dolist (pereche a-lista1 rezultat)
         (if (not (equal (car pereche) poz-blanc1))
            (let ((poz1 (car pereche)) 
                  (poz2 (car (rassoc (cdr pereche) a-lista2))))
               ;;(print rezultat)
               (setf rezultat (+ rezultat
                                 (manhattan poz1 poz2)))
               )))))
       
(defun manhattan (poz1 poz2)
   (+ (abs (- (floor (1- poz1) 3) (floor (1- poz2) 3)))
      (abs (- (mod (1- poz1) 3) (mod (1- poz2) 3)))))   

;;Exemplu de apel
(cautcale (list (make-cale :stare *INITIAL* :g 0 :h 0 :f 0)) 
  #'estescop* #'succf* '() #'test-st* #'h1 #'g 0)

