;;/////////////////////////////////////////////////////////////////
;;
;; 4 Logica si demonstrarea teoremelor
;; 4.2 Unificarea expresiilor
;; 4.2.2 Implementarea unificarii
;;
;;/////////////////////////////////////////////////////////////////

;; "unificare.lsp" - implementarea algoritmului de unificare a termenilor

;; se defineste macrocaracterul ? astfel incit ?<simbol> se transforma 
;; intr-o lista (*var* <simbol>)
(set-macro-character #\?
  #'(lambda (stream char)
       (list '*var* (read stream t nil t))))

(defun variabila-p (var)
  "Intoarce t daca argumentul este o variabila si nil in caz contrar."
  (and (listp var) (eq (first var) '*var*)))

(defun extrage-substitutie (variabila mgu)
  "Intoarce substitutia variabilei in unificatorul mgu."
  (cdr (assoc variabila mgu :test #'equal)))

(defun adauga-substitutie (variabila termen   					mgu)
  "Adauga o substitutie variabilei in unificatorul mgu."
  (cons (cons variabila termen) mgu))

(defun apare-in-p (variabila termen mgu)
  "Intoarce t daca termenul contine variabila."
  (cond 
    ((atom termen) nil)
     ;;termen constant deci nu contine variabila
    ((variabila-p termen)
     (or (equal variabila termen)
         (apare-in-p variabila					   			(extrage-substitutie termen mgu) mgu)))
    (t (or (apare-in-p variabila (first termen) mgu)
           (apare-in-p variabila (rest termen) mgu)))))

(defun unifica (TA TB &optional mgu)
  "Verifica daca termenii TA si TB unifica si construieste cel mai general unificator mgu."
  (cond ((variabila-p TA)
         (unifica-variabila TA TB mgu))
        ((variabila-p TB)
         (unifica-variabila TB TA mgu))
        ((atom TA)
         (when (eq TA TB)
	        ;; constante identice
              (values t mgu)))								;; nu se adauga o substitutie noua
        ((atom TB) nil)
		;; termenii nu unifica
        ((and (eq (first TA) (first TB))
		;; termeni functionali
              (= (length TA) (length TB)))
		;; au acelasi functor si aritate?
         (let (succes)
          (when (every #'(lambda (TAi TBi)
			    	(multiple-value-setq
				   (succes mgu)
				   (unifica TAi TBi mgu)))
                   (rest TA) (rest TB))
            (values t mgu))))))

(defun unifica-variabila (var term mgu)
  "Unifica variabila cu termenul term in prezenta unificatorului partial mgu."
  (if (equal var term)
		;;variabila unifica cu ea insasi?
    (values t mgu)
		;; dac da, nu adauga substitutia
    (let ((S (extrage-substitutie var mgu)))
      (cond (S  (unifica S term mgu))
		;; variabila are deja o substitutie
            ((not (apare-in-p var term mgu))
             (values t (adauga-substitutie var term  
                           mgu)))))))

(defun substituie-variabile (termen mgu)
  "Intoarce termenul in care variabilele au fost substituite."
  (cond 
   ((atom termen) termen)
	;; termen constant
   ((variabila-p termen)
	;; termen variabil
    (let ((S (extrage-substitutie termen mgu)))
      (if S
		;; daca are o substitutie,
        (substituie-variabile S mgu)
		;; verifica substitutia 
        termen)))
		;; altfel intoarce termenul

   (t (cons 
       (substituie-variabile (first termen) mgu)
       (substituie-variabile (rest termen) mgu)))))

;;Exemple de apel
(unifica '?x 'a '())
(unifica '(P ?x ?y) '(P (f a b) ?z))
(unifica '(P (f ?x a) (g ?x b)) '(P (f b a) (g ?u b)))

(defun termeni-unificati (t1 t2)
   (let (rez mgu) 
      (multiple-value-setq (rez mgu) (unifica t1 t2))
      (if rez
         (substituie-variabile t1 mgu))))

;;Exemple de apel
(termeni-unificati '(P ?x ?y) '(P (f a b) ?z))
(termeni-unificati '(P (f ?x a) (g ?x b)) '(P (f b a) (g ?u b)))